#ifndef CoCoA_NumTheory_H
#define CoCoA_NumTheory_H

//   Copyright (c)  2009  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.
//   -------------------------------------------------------------------------

#include "CoCoA/MachineInteger.H"
#include "CoCoA/factorization.H"
#include "CoCoA/ZZ.H"

namespace CoCoA
{

  unsigned long PowerMod(const MachineInteger& base, const MachineInteger& exp, const MachineInteger& modulus);
  ZZ PowerMod(const ZZ& base, const MachineInteger& exp, const ZZ& modulus);
  ZZ PowerMod(const MachineInteger& base, const ZZ& exp, const ZZ& modulus);
  ZZ PowerMod(const ZZ& base, const ZZ& exp, const ZZ& modulus);
  // other cases???

  unsigned long gcd(const MachineInteger& a, const MachineInteger& b);
  const ZZ::rtn gcd(const ZZ& A, const MachineInteger& b);
  const ZZ::rtn gcd(const MachineInteger& a, const ZZ& B);
  const ZZ::rtn gcd(const ZZ& A, const ZZ& B);

  unsigned long ExtGcd(long& CofacA, long& CofacB, const MachineInteger& a, const MachineInteger& b);
//???? signature???  ZZ ExtGcd(ZZ& CoFacA, ZZ& CofacB, const MachineInteger& A, const ZZ& B);
//???? signature???  ZZ ExtGcd(ZZ& CoFacA, ZZ& CofacB, const ZZ& A, const MachineInteger& B);
  ZZ ExtGcd(ZZ& CoFacA, ZZ& CofacB, const ZZ& A, const ZZ& B);

  unsigned long InvMod(const MachineInteger& r, const MachineInteger& m);
  unsigned long InvMod(const ZZ& r, const MachineInteger& m);
  ZZ InvMod(const MachineInteger& r, const ZZ& m);
  ZZ InvMod(const ZZ& r, const ZZ& m);

  unsigned long lcm(const MachineInteger& a, const MachineInteger& b);
  const ZZ::rtn lcm(const ZZ& A, const MachineInteger& b);
  const ZZ::rtn lcm(const MachineInteger& a, const ZZ& B);
  const ZZ::rtn lcm(const ZZ& a, const ZZ& b);


  bool IsPrime(const MachineInteger& n);
  bool IsPrime(const ZZ& N); // might be very slow!!
  bool IsProbPrime(const MachineInteger& n, unsigned int NumIters = 25); // probably not very useful
  bool IsProbPrime(const ZZ& N, unsigned int NumIters = 25);

  unsigned long NextPrime(const MachineInteger& n); ///< next prime number after n or 0 if there is none, must have n>0
  unsigned long PrevPrime(const MachineInteger& n); ///< previous prime number before n or 0 if there is none, must have n>0

  // No NextPrime for ZZ, use NextProbPrime and call IsPrime if you want a (costly) guarantee of primality.
  ZZ::rtn NextProbPrime(ZZ N, unsigned int NumIters = 25); ///< next probable prime number after abs(N)
  ZZ::rtn PrevProbPrime(ZZ N, unsigned int NumIters = 25); ///< previous probable prime number before abs(N) or 0 if there is none

// struct PrimeSourceStruct
// {
//   FFelem start, p;
// };
// typedef struct PrimeSourceStruct *PrimeSource;

// PrimeSource PrimeSourceCtor(void);
// void PrimeSourceDtor(PrimeSource PS);
// FFelem NextPrime(PrimeSource PS);

  factorization<long> SmoothFactor(const MachineInteger& N, const MachineInteger& TrialLimit);
//????  factorization<unsigned long> SmoothFactor(MachineInteger N, MachineInteger TrialLimit);
  factorization<ZZ>   SmoothFactor(const ZZ& N, const MachineInteger& TrialLimit);
  factorization<ZZ>   SmoothFactor(const ZZ& N, const ZZ& TrialLimit);

  factorization<long> factor(const MachineInteger& N);
  factorization<ZZ>   factor(const ZZ& N);

  unsigned long EulerPhi(const MachineInteger& n);
  ZZ EulerPhi(const ZZ& N);

  int PrimitiveRoot(const MachineInteger& p);

} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/NumTheory.H,v 1.1 2009/06/05 12:14:55 abbott Exp $
// $Log: NumTheory.H,v $
// Revision 1.1  2009/06/05 12:14:55  abbott
// Major change:
//   created new files NumTheory.H/C  which contain basic number theory operations
//   removed several basic number theory operations from ZZ.H/C
//   removed gcd from MachineInteger.H/C
//   changed names of some basic fns:
//      IsPPrime -> IsProbPrime
//      invmod -> InvMod    (changed signature too)
//      powermod -> PowerMod  (changed signature too)
//   added new fns
//      NextProbPrime & PrevProbPrime
//   consequent changes to other code and tests and examples
//
//

#endif
