#ifndef CoCoA_convert_H
#define CoCoA_convert_H

//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.
//   -------------------------------------------------------------------------

#include "CoCoA/error.H"

#include <string>
// using std::string;

namespace CoCoA
{

  class ZZ; // fwd decl, defined in ZZ.H

  bool convert(long& lhs, const ZZ& src);
  bool convert(int& lhs, const ZZ& src);
  bool convert(unsigned long& lhs, const ZZ& src);
  bool convert(unsigned int& lhs, const ZZ& src);
  bool convert(std::string& lhs, const ZZ& src, int base=10);
  bool convert(ZZ& lhs, const std::string& src, int base=0);

  bool convert(ZZ& num, ZZ& den, double z);   ///< convert double z into an exact rational ==> den will be a power of 2.
  bool convert(double& z, const ZZ& N);                   ///< conversion may discard information!
  bool convert(double& z, const ZZ& num, const ZZ& den);  ///< conversion may discard information!



  // Convert various sorts of unsigned integral value into a signed long;
  // result is true iff faithful conversion is possible.
  // All of these functions employ a static_cast to do the real work.
  // One of these functions will also convert a std::size_t to a long.
  inline bool convert(long& dest, const unsigned long& src)
  { dest = static_cast<long>(src); return (dest >= 0); }

  inline bool convert(long& dest, const unsigned int& src)
  { dest = static_cast<long>(src); return (dest >= 0); }

  inline bool convert(long& dest, const unsigned short& src)
  { dest = static_cast<long>(src); return true; } // always succeeds

  inline bool convert(long& dest, const unsigned char& src)
  { dest = static_cast<long>(src); return true; } // always succeeds


  // Write it this way to allow user to specify just one of the template args when calling.
  template<typename T>
  class CheckedCast
  {
  public:
    template<typename T2>
    CheckedCast(const T2& val, const char* const FnName) { if (!convert(myVal, val)) CoCoA_ERROR(ERR::BadCheckedCast, FnName); }
    operator T() const { return myVal; }
  private: // disable copy ctor and assignment
    CheckedCast(const CheckedCast<T>&);                // NEVER DEFINED -- copy ctor disabled
    CheckedCast<T>& operator=(const CheckedCast<T>&);  // NEVER DEFINED -- assignment disabled
  private: // data member
    T myVal;
  };


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/convert.H,v 1.3 2008/07/21 07:59:11 abbott Exp $
// $Log: convert.H,v $
// Revision 1.3  2008/07/21 07:59:11  abbott
// Improved two comments.
//
// Revision 1.2  2007/10/30 17:14:11  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1  2007/03/23 18:38:43  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
//

#endif
