//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/CanonicalHom.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/FractionField.H"

//#include <iostream>
using std::ostream;


namespace CoCoA
{

  RingHom CanonicalHom(const ring& domain, const ring& codomain)
  {
    if (domain == codomain) return IdentityHom(domain);
    // check codomain
    if (IsFractionField(codomain))
    {
      FractionField FrF = AsFractionField(codomain);
      if (domain == BaseRing(FrF))
        return EmbeddingHom(FrF);
      goto CheckDomain;
    }
    if (IsPolyRing(codomain))
    {
      PolyRing P = AsPolyRing(codomain);
      if (domain == CoeffRing(P))
        return CoeffEmbeddingHom(P);
      goto CheckDomain;
    }
    if (IsQuotientRing(codomain))
    {
      QuotientRing QR = AsQuotientRing(codomain);
      if (domain == BaseRing(QR))
        return QuotientingHom(QR);
      goto CheckDomain;
    }
  CheckDomain:
    if (IsZ(domain)) return ZEmbeddingHom(codomain);
    if (IsQ(domain)) return QEmbeddingHom(codomain);
    CoCoA_ERROR(ERR::CanonicalHomFail, "CanonicalHom(R1,R2)");
    return IdentityHom(codomain); // Never executed; just to keep the compiler quiet.
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/CanonicalHom.C,v 1.2 2007/10/30 17:14:08 abbott Exp $
// $Log: CanonicalHom.C,v $
// Revision 1.2  2007/10/30 17:14:08  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.2  2007/03/07 11:18:01  bigatti
// -- fixed CanonicalHom goto flag
//
// Revision 1.1  2007/03/05 21:25:02  cocoa
// New CanonicalHom pseudo-ctor.
//
//
