//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/RingTwinFloat.H"
#include "CoCoA/io.H"

using namespace CoCoA;

#include <iostream>
using std::endl;

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


// This test unearthed a bug in GMP version 4.1.3.
// The test passes fine with GMP 4.1.4, and you are strongly
// encouraged to make sure you have this latter version or a
// still later version.

void program()
{
  GlobalManager CoCoAFoundations;

  // cout << "Program to check whether 1+delta differs from 1 for various small delta\n"
  //       << "and using various precisions of RingTwinFloat values." << endl;
  for (int prec=32; prec <= 1024; prec *= 2)
  {
    RingTwinFloat R = NewRingTwinFloat(prec);
    RingElem x(R,1);
    RingElem delta(R,1);

    try
    {
      for (int i=0; i < 999; ++i)
      {
        delta /= 2;
        TEST_ASSERT(x + delta != x);
        TEST_ASSERT(x - delta != x);
      }
      TEST_ASSERT(PrecisionBits(R) > 1000);
    }
    catch (const RingTwinFloat::InsufficientPrecision&)
    {
      TEST_ASSERT(PrecisionBits(R) < 1000);
    }
  }
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
