//   Copyright (c)  2007  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/convert.H"
#include "CoCoA/ZZ.H"
#include "CoCoA/io.H"

using namespace CoCoA;

#include <limits>
using std::numeric_limits;
#include <string>
using std::string;
#include <iostream>
using std::cin;
using std::cout;
using std::endl;

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


// VERY INCOMPLETE

void program()
{
  GlobalManager CoCoAFoundations;

  long l;
  unsigned long ul = numeric_limits<unsigned long>::max();
  int i = numeric_limits<int>::max();
  unsigned int ui = numeric_limits<unsigned int>::max();
  unsigned short us = numeric_limits<unsigned short>::max();
  unsigned char uc = numeric_limits<unsigned char>::max();

  TEST_ASSERT(convert(l, uc) && l == long(uc));
  TEST_ASSERT(convert(l, us) && l == long(us));
  if (double(numeric_limits<unsigned int>::max()) > double(numeric_limits<long>::max()))
    TEST_ASSERT(!convert(l, ui));
  else
  {
    TEST_ASSERT(convert(l, ui) && long(ui) == l);
  }
  unsigned int ui2 = numeric_limits<int>::max(); // biggest int surely fits into unsigned int
  TEST_ASSERT(convert(l, ui2) && l == long(ui2));
  TEST_ASSERT(!convert(l, ul));

  ZZ N;
  N = numeric_limits<long>::max();
  TEST_ASSERT(convert(l, N) && l == N);
  TEST_ASSERT(!convert(l, N+1));
  N = numeric_limits<long>::min();
  TEST_ASSERT(convert(l, N) && l == N);
  TEST_ASSERT(!convert(l, N-1));

  N = numeric_limits<unsigned long>::max();
  TEST_ASSERT(convert(ul, N) && ul == N);
  TEST_ASSERT(!convert(ul, N+1));
  N = numeric_limits<unsigned long>::min(); // just zero
  TEST_ASSERT(convert(ul, N) && ul == N);
  TEST_ASSERT(!convert(ul, N-1));

  N = numeric_limits<int>::max();
  TEST_ASSERT(convert(i, N) && i == N);
  TEST_ASSERT(!convert(i, N+1));
  N = numeric_limits<int>::min();
  TEST_ASSERT(convert(i, N) && i == N);
  TEST_ASSERT(!convert(i, N-1));

  N = numeric_limits<unsigned int>::max();
  TEST_ASSERT(convert(ui, N) && ui == N);
  TEST_ASSERT(!convert(ui, N+1));
  N = numeric_limits<unsigned int>::min(); // just zero
  TEST_ASSERT(convert(ui, N) && ui == N);
  TEST_ASSERT(!convert(ui, N-1));

  string s1, s2;
  s1 = "100000000000000000000000000000000000000000000000000";  // "1" followed by 50 zeros
  TEST_ASSERT(convert(N, s1) && N == power(10, 50));
  TEST_ASSERT(convert(s2, N) && s1 == s2);
  TEST_ASSERT(convert(N, s1, 10) && N == power(10, 50));
  TEST_ASSERT(convert(N, s1, 2) && N == power(2, 50));   // regard string as binary
  TEST_ASSERT(convert(s2, N, 2) && s1 == s2);            // produce binary string
  TEST_ASSERT(convert(N, s1, 13) && N == power(13, 50)); // regard string as base 13
  TEST_ASSERT(convert(s2, N, 13) && s1 == s2);           // produce base 13 string
  s1 = "-100000000000000000000000000000000000000000000000000"; // "-1" followed by 50 zeros
  TEST_ASSERT(convert(N, s1) && N == -power(10, 50));
  TEST_ASSERT(convert(s2, N) && s1 == s2);
  TEST_ASSERT(convert(N, s1, 10) && N == -power(10, 50));
  TEST_ASSERT(convert(N, s1, 2) && N == -power(2, 50));   // regard string as binary
  TEST_ASSERT(convert(s2, N, 2) && s1 == s2);             // produce binary string
  TEST_ASSERT(convert(N, s1, 13) && N == -power(13, 50)); // regard string as base 13
  TEST_ASSERT(convert(s2, N, 13) && s1 == s2);            // produce base 13 string


}

// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
