//   Copyright (c)  2009 Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BuildInfo.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/error.H"
#include "CoCoA/io.H"

#ifdef CoCoA_WITH_FROBBY
#include "CoCoA/RingQ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/TmpFrobby.H"
#include "CoCoA/ideal.H"
#endif

using namespace CoCoA;

#include <iostream>
using std::endl;
#ifdef CoCoA_WITH_FROBBY
#include <vector>
using std::vector;
#endif
//----------------------------------------------------------------------
// First test for Frobby library.
// Trivial computation to test proper integration.
//----------------------------------------------------------------------

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void program()
{
  GlobalManager CoCoAFoundations;

#ifdef CoCoA_WITH_FROBBY
  PolyRing P = NewPolyRing(RingQ(), 3);
  RingElem x(indet(P,0));
  RingElem y(indet(P,1));
  RingElem z(indet(P,2));
  
  //I := Ideal(x^2, x*y, y^2, z^2);
  vector<RingElem> g;
  g.push_back(x*x);
  g.push_back(x*y);
  g.push_back(y*y);
  g.push_back(z*z);

  ideal I = ideal(P, g);
  
  ideal AD = AlexanderDual(I, LPP(x*x*y*y*z*z));// = Ideal(x^2yz, xy^2z);
  TEST_ASSERT(AD == ideal(P, x*x*y*z) + ideal(P, x*y*y*z));

  ideal MSM = MaximalStandardMonomials(I);// = Ideal(yz, xz);
  TEST_ASSERT(MSM == ideal(P, y*z) + ideal(P, x*z));

  vector<ideal> ID;
  IrreducibleDecomposition(I, ID);// = [Ideal(x, y^2, z^2), Ideal(x^2, y, z^2)];
  TEST_ASSERT(ID[0] == ideal(P, x) + ideal(P, y*y) + ideal(P, z*z));
  TEST_ASSERT(ID[1] == ideal(P, x*x) + ideal(P, y) + ideal(P, z*z));
#endif
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/tests/test-frobby1.C,v 1.1 2009/02/11 15:08:26 bigatti Exp $
// $Log: test-frobby1.C,v $
// Revision 1.1  2009/02/11 15:08:26  bigatti
// -- first import
//

